package org.mandas.docker.client.messages.swarm;

import com.fasterxml.jackson.annotation.JsonProperty;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 * Immutable implementation of {@link Spread}.
 * <p>
 * Use the builder to create immutable instances:
 * {@code ImmutableSpread.builder()}.
 */
@SuppressWarnings({"all"})
final class ImmutableSpread implements Spread {
  private final String spreadDescriptor;

  private ImmutableSpread(String spreadDescriptor) {
    this.spreadDescriptor = spreadDescriptor;
  }

  /**
   * @return The value of the {@code spreadDescriptor} attribute
   */
  @JsonProperty("SpreadDescriptor")
  @Override
  public String spreadDescriptor() {
    return spreadDescriptor;
  }

  /**
   * Copy the current immutable object by setting a value for the {@link Spread#spreadDescriptor() spreadDescriptor} attribute.
   * An equals check used to prevent copying of the same value by returning {@code this}.
   * @param value A new value for spreadDescriptor
   * @return A modified copy of the {@code this} object
   */
  public final ImmutableSpread withSpreadDescriptor(String value) {
    String newValue = Objects.requireNonNull(value, "spreadDescriptor");
    if (this.spreadDescriptor.equals(newValue)) return this;
    return new ImmutableSpread(newValue);
  }

  /**
   * This instance is equal to all instances of {@code ImmutableSpread} that have equal attribute values.
   * @return {@code true} if {@code this} is equal to {@code another} instance
   */
  @Override
  public boolean equals(Object another) {
    if (this == another) return true;
    return another instanceof ImmutableSpread
        && equalTo((ImmutableSpread) another);
  }

  private boolean equalTo(ImmutableSpread another) {
    return spreadDescriptor.equals(another.spreadDescriptor);
  }

  /**
   * Computes a hash code from attributes: {@code spreadDescriptor}.
   * @return hashCode value
   */
  @Override
  public int hashCode() {
    int h = 5381;
    h += (h << 5) + spreadDescriptor.hashCode();
    return h;
  }

  /**
   * Prints the immutable value {@code Spread} with attribute values.
   * @return A string representation of the value
   */
  @Override
  public String toString() {
    return "Spread{"
        + "spreadDescriptor=" + spreadDescriptor
        + "}";
  }

  /**
   * Creates an immutable copy of a {@link Spread} value.
   * Uses accessors to get values to initialize the new immutable instance.
   * If an instance is already immutable, it is returned as is.
   * @param instance The instance to copy
   * @return A copied immutable Spread instance
   */
  public static ImmutableSpread copyOf(Spread instance) {
    if (instance instanceof ImmutableSpread) {
      return (ImmutableSpread) instance;
    }
    return ImmutableSpread.builder()
        .from(instance)
        .build();
  }

  /**
   * Creates a builder for {@link ImmutableSpread ImmutableSpread}.
   * <pre>
   * ImmutableSpread.builder()
   *    .spreadDescriptor(String) // required {@link Spread#spreadDescriptor() spreadDescriptor}
   *    .build();
   * </pre>
   * @return A new ImmutableSpread builder
   */
  public static ImmutableSpread.Builder builder() {
    return new ImmutableSpread.Builder();
  }

  /**
   * Builds instances of type {@link ImmutableSpread ImmutableSpread}.
   * Initialize attributes and then invoke the {@link #build()} method to create an
   * immutable instance.
   * <p><em>{@code Builder} is not thread-safe and generally should not be stored in a field or collection,
   * but instead used immediately to create instances.</em>
   */
  static final class Builder {
    private static final long INIT_BIT_SPREAD_DESCRIPTOR = 0x1L;
    private long initBits = 0x1L;

    private String spreadDescriptor;

    private Builder() {
    }

    /**
     * Fill a builder with attribute values from the provided {@code Spread} instance.
     * Regular attribute values will be replaced with those from the given instance.
     * Absent optional values will not replace present values.
     * @param instance The instance from which to copy values
     * @return {@code this} builder for use in a chained invocation
     */
    public final Builder from(Spread instance) {
      Objects.requireNonNull(instance, "instance");
      spreadDescriptor(instance.spreadDescriptor());
      return this;
    }

    /**
     * Initializes the value for the {@link Spread#spreadDescriptor() spreadDescriptor} attribute.
     * @param spreadDescriptor The value for spreadDescriptor 
     * @return {@code this} builder for use in a chained invocation
     */
    @JsonProperty("SpreadDescriptor")
    public final Builder spreadDescriptor(String spreadDescriptor) {
      this.spreadDescriptor = Objects.requireNonNull(spreadDescriptor, "spreadDescriptor");
      initBits &= ~INIT_BIT_SPREAD_DESCRIPTOR;
      return this;
    }

    /**
     * Builds a new {@link ImmutableSpread ImmutableSpread}.
     * @return An immutable instance of Spread
     * @throws java.lang.IllegalStateException if any required attributes are missing
     */
    public ImmutableSpread build() {
      if (initBits != 0) {
        throw new IllegalStateException(formatRequiredAttributesMessage());
      }
      return new ImmutableSpread(spreadDescriptor);
    }

    private String formatRequiredAttributesMessage() {
      List<String> attributes = new ArrayList<>();
      if ((initBits & INIT_BIT_SPREAD_DESCRIPTOR) != 0) attributes.add("spreadDescriptor");
      return "Cannot build Spread, some of required attributes are not set " + attributes;
    }
  }
}
